import time
from multiprocessing import Process, Queue

import GPy
from sklearn.preprocessing import MinMaxScaler
import numpy as np


def predict_all(m: GPy.models.GPCoregionalizedRegression, X):
    ny = len(np.unique(m.output_index))
    y = []
    covy = []
    Xaug = np.hstack((X, 0.0 * np.ones_like(X[:, 0:1])))
    for iy in range(ny):
        Xaug[:, -1:] = iy
        y_i, covy_i = m.predict(Xaug, Y_metadata={'output_index': Xaug[:, -1:].astype(int)})
        y.append(y_i)
        covy.append(covy_i)
    return np.hstack(y), np.hstack(covy)

class GPModel(Process):
    def __init__(self):
        super(GPModel, self).__init__()
        self.train_queue = Queue()
        self.test_queue = Queue()
        self.output = Queue()

        self.norm_x = None
        self.norm_y = None
        self.normalize_y = False
        self.m = None

    def run(self) -> None:

        X_train, Y_train, num_restarts, normalize_y = self.train_queue.get()
        np.random.seed(2311)

        self.norm_x = MinMaxScaler((0, 1))
        X_train = self.norm_x.fit_transform(X_train)

        self.normalize_y = normalize_y
        if self.normalize_y:
            self.norm_y = MinMaxScaler((-1, 1))
            Y_train = self.norm_y.fit_transform(Y_train)
            # print(norm_y.min_)
            # print(norm_y.scale_)
            # print(norm_y.feature_range)
        # print(X_train.shape)
        Y_train = np.array(list(zip(*Y_train)))
        Y_train = np.array([i[:, None] for i in Y_train])
        K = GPy.kern.RBF(input_dim=X_train.shape[1])
        # print(Y_train.shape)
        icm = GPy.util.multioutput.ICM(input_dim=X_train.shape[1], num_outputs=Y_train.shape[1], kernel=K)
        # print(np.array)
        self.m = GPy.models.GPCoregionalizedRegression([X_train, X_train], Y_train, kernel=icm)

        if num_restarts:
            # self.m.optimize(messages=True)
            self.m.optimize_restarts(messages=True, num_restarts=num_restarts)
        print(self.m)
        self.output.put(None)
        while True:
            X_test = self.test_queue.get()

            if X_test is None:
                return

            X_test = self.norm_x.transform(X_test)

            u_predicted, u_covariance = predict_all(self.m, X_test)

            if self.normalize_y:
                u_predicted = self.norm_y.inverse_transform(u_predicted)
                u_covariance /= self.norm_y.scale_ ** 2

            self.output.put((u_predicted, u_covariance, str(self.m)))

    def train(self, X_train, Y_train, num_restarts=0, normalize_y=False):
        self.train_queue.put((X_train, Y_train, num_restarts, normalize_y))
        return self.output.get()

    def test(self, X_test):
        self.test_queue.put(X_test)
        return self.output.get()

    def exit(self):
        self.test_queue.put(None)

