import numpy as np
import pandas as pd

import hex3_chen_old as hex3_chen
from g_process import GPModel
from optimal_u.hex3_gen_u_optim import load_data


def measurements_to_array(u, meas, d_order):
    # Distubances
    if isinstance(meas, pd.DataFrame):
        meas = meas.to_dict('records')

    if isinstance(u, pd.DataFrame):
        u_u0 = u['u0']
        u_u1 = u['u1']
    else:
        u_u0 = u[0]
        u_u1 = u[1]

    # Ensure keys keep order
    if d_order is None:
        d_order = list(d[0].keys())
        # Remove the heat loss. Otherwise, random noise will be applied to 0 values, degrades performance.
        d_order.remove('h1')
        d_order.remove('h2')
        d_order.remove('h3')
        d_order.remove('Ts')

    d_keys = d_order.copy()
    toggle = 'alpha1' in d_keys

    if toggle:
        d_keys.remove('alpha1')
        d_keys.remove('alpha2')

    # Inputs
    if isinstance(meas, np.ndarray):
        if toggle:
            # Move
            X = np.array([np.append(np.array([u_u0, u_u1]), meas[:-2])])
        else:
            X = np.array([meas])
    else:
        X = [*([u_u0, u_u1] * toggle), *[meas[k] for k in d_keys]]
        X = np.array([X], dtype=np.float)
    return X


def disturbance_to_array(u, d, d_order):
    # Distubances
    if isinstance(d, pd.DataFrame):
        d = d.to_dict('records')

    # Optimal u values

    # This is u_chen
    if isinstance(u, pd.DataFrame):
        u_u0 = u['u0']
        u_u1 = u['u1']
    else:
        u_u0 = u[0]
        u_u1 = u[1]

    # Ensure keys keep order
    if d_order is None:
        d_order = list(d[0].keys())
        # Remove the heat loss. Otherwise, random noise will be applied to 0 values, degrades performance.
        d_order.remove('h1')
        d_order.remove('h2')
        d_order.remove('h3')
        d_order.remove('Ts')

    d_keys = d_order.copy()
    toggle = 'alpha1' in d_keys

    if toggle:
        d_keys.remove('alpha1')
        d_keys.remove('alpha2')

    # Inputs
    X = [*([u_u0, u_u1] * toggle), *[d[k] for k in d_keys]]
    X = np.array([X], dtype=np.float)
    return X


def preprocess(u: pd.DataFrame, d: [pd.DataFrame, np.ndarray], d_order=None, u_input=None, add_noise=False):
    # Distubances
    if isinstance(d, pd.DataFrame):
        d = d.to_dict('records')

    # Optimal u values

    # This is u_chen
    if isinstance(u, pd.DataFrame):
        u_u0 = u['uc0']
        u_u1 = u['uc1']
    else:
        u_u0 = u[:, 0]
        u_u1 = u[:, 1]

    X = []

    # Ensure keys keep order
    if d_order is None:
        d_order = list(d[0].keys())
        # Remove the heat loss. Otherwise, random noise will be applied to 0 values, degrades performance.
        d_order.remove('h1')
        d_order.remove('h2')
        d_order.remove('h3')
        d_order.remove('Ts')

    # Random valve inputs
    if isinstance(u_input, pd.DataFrame):
        u_in0 = u_input['ur0']
        u_in1 = u_input['ur1']
    elif isinstance(u_input, np.ndarray):
        u_in0 = u_input[:, 0]
        u_in1 = u_input[:, 1]
    else:
        raise Exception('This should not be reached')

    d_keys = d_order.copy()
    toggle = 'alpha1' in d_keys

    if toggle:
        d_keys.remove('alpha1')
        d_keys.remove('alpha2')
    # Inputs

    for idx, (u0, u1, d_row) in enumerate(zip(u_in0, u_in1, d)):
        x = [*([u0, u1] * toggle), *[d_row[k] + add_noise * np.random.normal(0, 1) for k in d_keys]]
        X.append(x)

    X = np.array(X)
    # Optimal u values
    Y = np.array(list(zip(u_u0, u_u1)))

    return X, Y, d_order


def generate_meas_set_data(mset, u, d):
    d = d.to_dict('records')

    if isinstance(u, pd.DataFrame):
        u = np.array(list(zip(u['ur0'], u['ur1'])))

    results = []

    for idx, (u_, d_) in enumerate(zip(u, d)):
        # print(u_)
        # print(d_)
        result = hex3_chen.output_meas(mset, u_, d_)
        if not result['success']:
            print('Failed, small step in u')
            print(u_)

        res_dict = {}
        for idx, key in enumerate(hex3_chen.meas_sets[mset]):
            res_dict[key] = result['y'][idx]
        results.append(res_dict)
    return np.array(results, dtype=dict)


class InputController:
    def __init__(self, training_database, meas_set=None, noise=False):
        train_data = load_data(training_database)

        if meas_set is not None:
            d_keys = hex3_chen.meas_sets[meas_set]
            ur = train_data['u_rand']
            uc = train_data['u_chen']
            u1 = np.array(list(zip(ur['ur0'], ur['ur1'])))
            u2 = np.array(list(zip(uc['uc0'], uc['uc1'])))
            diff = u2 - u1
            split = int(0.30 * len(u1))  # 0.7 means 30% non-optimal
            u3 = np.concatenate((u1[:split] + diff[:split], u1[split:] + diff[split:] * 0.5))
            # u3 = u1  # only random
            # u3 = u2  # only optimal
            disturbances_train = generate_meas_set_data(meas_set, u3, train_data['d'])

        else:
            raise Exception('Not supported for this controller')
            # d_keys = None
            # disturbances_train = train_data['d']
            # u3 = train_data['u_rand']

        X_train, Y_train, key_order = preprocess(train_data['u_chen'], disturbances_train, d_order=d_keys,
                                                 u_input=u3, add_noise=noise)
        # print(X_train[0])
        # print(Y_train[0])
        self.outputs = len(Y_train[0])
        self.real_cv = np.empty((0, 2), int)
        self.t_samples = len(Y_train)
        self.meas_set = meas_set
        self.d_keys = key_order
        self.gp = GPModel()
        self.gp.start()
        self.gp.train(X_train, Y_train, num_restarts=1, normalize_y=True)

    def __str__(self):
        return 'u'

    def plot_init(self, axes):
        lines = []
        rlines = []
        for idx, ax in enumerate(axes):
            ax.set_ylabel(f'Valve {idx + 1} [-]')
            ax.set_xlabel('Time [-]')
            ax.set_ylim([0.25, 0.35])
            ax.set_xlim([0, 1])
            line, = ax.plot([], [], label='Predicted')
            lines.append(line)
            rline, = ax.plot([], [], '--', label='Optimal')
            rlines.append(rline)
            ax.legend()
        return lines, rlines

    def gen_real_cv(self, u, u_opt, d):
        self.real_cv = np.append(self.real_cv, np.array([u_opt]), axis=0)

    def plot_prediction(self, axes, lines, time, variables, real_lines):
        for idx, (ax, line) in enumerate(zip(axes, lines)):
            ylim = ax.get_ylim()
            # print(ylim)
            if not (ylim[0] <= variables[-1][idx] <= ylim[1]):
                ax.set_ylim([np.min(variables[:, idx])-0.05, np.max(variables[:, idx])+0.05])
            ylim = ax.get_ylim()
            if not (ylim[0] <= self.real_cv[-1][idx] <= ylim[1]):
                ax.set_ylim([np.min(self.real_cv[:, idx])-0.05, np.max(self.real_cv[:, idx])+0.05])

            # Fix for auto-plotting
            if self.meas_set == 3 and idx == 0:
                ax.set_ylim([ylim[0], 0.38])
            ax.set_xlim([0, len(time)])
            line.set_xdata(time)
            line.set_ydata(variables[:, idx])
            real_lines[idx].set_xdata(time)
            real_lines[idx].set_ydata(self.real_cv[:, idx])

    def predict(self, u, measurement):
        # print('.-'*30)
        # print(u, measurement)
        try:
            if self.meas_set is not None:
                X = measurements_to_array(u, measurement, self.d_keys)
            else:
                X = disturbance_to_array(u, measurement, self.d_keys)
                raise Exception('Unhandled')
                # print(X)
        except:
            self.close()
            raise
        u_old = u
        alpha = 0.95
        u, u_var, _ = self.gp.test(X)  # print(f'Before {u} | grad {gradients}')
        u = u[0, :]
        # print(u)
        return alpha*u + (1. - alpha)*np.array(u_old), alpha*u + (1. - alpha)*np.array(u_old)

    def __del__(self):
        self.close()

    def close(self):
        self.gp.exit()


if __name__ == '__main__':
    # train = f'..\\gradient\\datasets\\train_gradient{500}.csv'
    # controller = GradientController(train)
    # controller.close()
    # print('Controller closed')
    pass
