import enum
import code
import matplotlib
import numpy as np
from matplotlib.backend_bases import MouseButton

import hex3_chen_old as hex3_chen

matplotlib.use('Qt5agg')
import matplotlib.pyplot as plt


class state(enum.IntEnum):
    alpha3 = 0
    T = 1
    Tstar1 = 2
    Tstar2 = 3
    Tstar3 = 4
    The1 = 5
    The2 = 6
    The3 = 7
    Q1 = 8
    Q2 = 9
    Q3 = 10
    Qloss1 = 11
    Qloss2 = 12
    Qloss3 = 13
    T1 = 14
    T2 = 15
    T3 = 16


def apply_noise(x, MS):
    noise = np.random.normal(0, 1, x.size)
    # No noise on valves
    if MS in (4, -4):
        noise[-2:] = 0

    return x + noise


def line_intersection(line1, line2):
    xdiff = (line1[0][0] - line1[1][0], line2[0][0] - line2[1][0])
    ydiff = (line1[0][1] - line1[1][1], line2[0][1] - line2[1][1])

    def det(a, b):
        return a[0] * b[1] - a[1] * b[0]

    div = det(xdiff, ydiff)
    if div == 0:
       raise Exception('lines do not intersect')

    d = (det(*line1), det(*line2))
    x = det(d, xdiff) / div
    y = det(d, ydiff) / div
    return x, y


zoomed = False
def on_click(event):
    """Enlarge or restore the selected axis."""
    global zoomed
    ax = event.inaxes

    if ax is None:
        # Occurs when a region not in an axis is clicked...
        return
    if event.button is MouseButton.LEFT:
        if zoomed:
            return
        # On left click, zoom the selected axes
        ax._orig_position = ax.get_position()
        ax.set_position([0.1, 0.1, 0.85, 0.85])
        for axis in event.canvas.figure.axes:
            # Hide all the other axes...
            if axis is not ax:
                axis.set_visible(False)
        zoomed = True
    elif event.button is MouseButton.RIGHT:
        # On right click, restore the axes
        zoomed = False
        try:
            ax.set_position(ax._orig_position)
            for axis in event.canvas.figure.axes:
                axis.set_visible(True)
        except AttributeError:
            # If we haven't zoomed, ignore...
            pass
    else:
        # No need to re-draw the canvas if it's not a left or right click
        return
    event.canvas.draw()
    event.canvas.flush_events()


# Setup system
def main(MS, do_plot, controller, disturb_table, noise):
    N = 500  # Min
    parameters = {}

    # Defining disturbance box [center, variability]
    parameters['T0'] = [60, 10]  # C
    parameters['w0'] = [105, 25]  # kW/K
    parameters['wh1'] = [40, 10]  # kW/K
    parameters['wh2'] = [50, 10]  # kW/K
    parameters['wh3'] = [30, 10]  # kW/K
    parameters['Th1'] = [150, 30]  # C
    parameters['Th2'] = [150, 30]  # C
    parameters['Th3'] = [150, 30]  # C
    parameters['UA1'] = [65, 15]  # kW/K
    parameters['UA2'] = [80, 10]  # kW/K
    parameters['UA3'] = [95, 15]  # kW/K

    parameters['Ts'] = [0, 0]  # C
    parameters['h1'] = [0, 0]  # kW/K
    parameters['h2'] = [0, 0]  # kW/K
    parameters['h3'] = [0, 0]  # kW/K

    d = {}
    for i, parname in enumerate(parameters.keys()):
        d[parname] = parameters[parname][0]

    u = np.array([0.3, 0.3])

    # Plotting initialization
    T_step = np.array([])  # T out
    T_opt_step = np.array([])  # T out optimal
    t_step = np.array([])  # Min, time step
    d_step = np.array([])
    x_step = []
    num_cv = controller.outputs
    u_step = np.empty((0, num_cv))
    u_opt_step = np.empty((0, num_cv))
    # Controlled variables
    cv_step = np.empty((0, num_cv))

    if do_plot:
        fig = plt.figure()
        controller_name = str(controller)

        title = f"Controller: {controller_name} | " \
                f"MS {MS} | Noise: {noise} | Samples {controller.t_samples} "
        # file_title = title.replace('|', '').replace(' ', '_')
        fig.suptitle(title)
        fig.canvas.mpl_connect('button_press_event', on_click)
        cost_ax = fig.add_subplot(221)
        state_T_line, = cost_ax.plot(t_step, T_step, label='State')
        opt_T_line, = cost_ax.plot(t_step, T_opt_step, '--', label='Optimal')
        cost_ax.set_ylabel('Cost [C]')
        cost_ax.set_xlabel('Time [-]')
        cost_ax.set_ylim([90, 135])
        cost_ax.set_xlim([0, 1])
        cost_ax.legend()

        cv_ax2 = fig.add_subplot(222)
        cv_ax2.set_ylim([110, 155])
        cv_ax2.set_xlim([0, 1])

        cv_axes = [cv_ax2]
        if num_cv == 2:
            cv_ax3 = fig.add_subplot(224)
            cv_ax3.set_ylim([110, 155])
            cv_ax3.set_xlim([0, 1])
            cv_axes.append(cv_ax3)
        elif num_cv > 2:
            raise Exception('Unhandled amount of controlled variables, for plotting.')

        cv_lines, real_cv_lines = controller.plot_init(cv_axes)

        # Integral cost plot
        loss_ax = fig.add_subplot(223)
        int_loss_line, = loss_ax.plot([], [])
        loss_ax.set_ylim([0, 10])
        loss_ax.set_xlim([0, 1])
        loss_ax.set_ylabel('Loss [C]')
        loss_ax.set_xlabel('Time [-]')

        # fig.set_dpi(160)
        # fig.show()
    max_step_loss = 0
    loss = 0
    neg_loss = 0
    old_loss = 0
    loss_step = np.array([])

    # Simulation loop

    # sys.stderr = open(os.devnull, "w")
    def disturb_system(d, t, parameters, pertub_table):
        changes = pertub_table.get(t, None)
        if changes is None:
            return d, None

        text = None
        for param, weight in changes:
            d[param] = parameters[param][0] + weight * parameters[param][1]
            if weight == 1.0:
                text = param
        return d, text

    for t in range(N):
        # introduce disturbance

        # TODO: Add convenient distrubance update function.
        #  Eg. by having list with [parameter, x% off nominal, time]
        # set to 0 to go back to nominal

        # 1. Th1 set 1, -1, 0, TH2....
        # 2. All 0.5
        # 3. Benchmark run

        # Tabels controllers vs loss from test

        # Maximum deviation error

        # Plot uncertainty

        d, plot_text = disturb_system(d, t, parameters, disturb_table)
        # print(d)
        x = hex3_chen.output_meas(MS, u, d)
        if noise:
            x['y'] = apply_noise(x['y'], MS)

        assert x['success']
        # print(x)
        # print(x['x'][state.T])
        # print(len(x['x']))
        x_opt = hex3_chen.optim(d)
        assert x_opt['success']
        # print(x_opt['x'][state.T])
        # print(x['y'])
        u_opt = x_opt['u']
        # if x['y'][hex3_chen.meas_sets[MS].index('T1')] > hex3_chen.Ti_max:
        #     print('Constraint violation')

        # Loss calculation
        # Only positive loss
        if t != 0:
            old_loss += ((x_opt['x'][state.T] - x['T']) + (T_opt_step[-1] - T_step[-1])) / 2
            # Cases:
            # 0. Constantly at the optimal
            # 1b. Lines not crossing, positive loss
            # 1a. Lines does not cross, pure negative loss.

            # 2. Lines cross, from positive to negative loss
            # 3. Lines cross, from negative to positive loss

            current_T, last_T = x['T'], T_step[-1]
            current_opt_T, last_opt_T = x_opt['x'][state.T], T_opt_step[-1]

            # Case 0:
            if current_T == current_opt_T and last_T == last_opt_T:
                pass
            # Case 1a:
            elif current_T <= current_opt_T and last_T <= last_opt_T:
                ls = ((current_opt_T - current_T) + (last_opt_T - last_T)) * 1. / 2
                assert ls >= 0
                loss += ls
            # Case 1b:
            elif current_T > current_opt_T and last_T > last_opt_T:
                ns = ((current_opt_T - current_T) + (last_opt_T - last_T)) * 1. / 2
                assert ns <= 0
                neg_loss += ns
            # Case 2: Lines cross, from positive to negative loss
            elif current_T >= current_opt_T and last_T < last_opt_T:
                T_line = ((t-1, last_T), (t, current_T))
                T_opt_line = ((t-1, last_opt_T), (t, current_opt_T))
                t_cross, T_cross = line_intersection(T_line, T_opt_line)
                ls = (0 + (last_opt_T - last_T)) * (t_cross - (t-1)) / 2
                ns = ((current_opt_T - current_T) + 0) * (t - t_cross) / 2

                try:
                    assert ls >= 0
                    assert ns <= 0
                except AssertionError:
                    print(f'Current T: {current_T} | OPT {current_opt_T}')
                    print(f'Last T: {last_T} | OPT {last_opt_T}')
                    print(f'ns: {current_T} | OPT {current_opt_T}')

                loss += ls
                neg_loss += ns

            # Case 3: Lines cross, from negative to positive loss
            elif current_T < current_opt_T and last_T >= last_opt_T:
                T_line = ((t - 1, last_T), (t, current_T))
                T_opt_line = ((t - 1, last_opt_T), (t, current_opt_T))
                t_cross, T_cross = line_intersection(T_line, T_opt_line)

                ns = (0 + (last_opt_T - last_T)) * (t_cross - (t - 1)) / 2
                ls = ((current_opt_T - current_T) + 0) * (t - t_cross) / 2
                try:
                    assert ls >= 0
                    assert ns <= 0
                except AssertionError:
                    print(f'Current T: {current_T} | OPT {current_opt_T}')
                    print(f'Last T: {last_T} | OPT {last_opt_T}')
                    print(f'ns: {current_T} | OPT {current_opt_T}')

                loss += ls
                neg_loss += ns
            else:
                raise Exception('Unknown line behaviour')
            # if x_opt['x'][state.T] - x['T'] <= 0:
            #     loss += 0
            # elif x_opt['x'][state.T]:
            #     pass
            # else:
            #     loss += ((x_opt['x'][state.T] - x['T']) + (T_opt_step[-1] - T_step[-1])) / 2
            # max_step_loss = max(abs(x_opt['x'][state.T] - x['T']), max_step_loss)
            # if (x_opt['x'][state.T] - x['T']) < 0:
            #     print('Too hot')

        u, cv = controller.predict(u, x['y'])
        # print(f'u {u}, meas {MS}')

        # u = u_opt
        # print(cv)
        # print(f'New u: {u} | new cv: {cv}')

        # Updating control
        x_step.append(x['x'])
        cv_step = np.append(cv_step, np.array([cv]), axis=0)
        d_step = np.append(d_step, np.array([d]), axis=0)
        u_step = np.append(u_step, np.array([u]), axis=0)
        u_opt_step = np.append(u_opt_step, np.array([u_opt]), axis=0)

        T_step = np.append(T_step, x['T'])
        T_opt_step = np.append(T_opt_step, x_opt['x'][state.T])
        t_step = np.append(t_step, t)
        loss_step = np.append(loss_step, loss)
        # print(cv_step)
        # Plotting

        if do_plot:

            cost_ax.set_xlim([0, len(t_step)])
            cost_ax.set_ylim([min(T_step)-2, max(T_step)+2])
            state_T_line.set_xdata(t_step)
            state_T_line.set_ydata(T_step)

            opt_T_line.set_xdata(t_step)
            opt_T_line.set_ydata(T_opt_step)

            controller.gen_real_cv(u, x_opt['u'], d)
            controller.plot_prediction(cv_axes, cv_lines, t_step, cv_step, real_cv_lines)

            if plot_text is not None:
                for idx, cv_ax in enumerate(cv_axes):

                    if cv_step[-1][idx] > controller.real_cv[-1][idx]:
                        if cv_step[-1][idx]-cv_step[-2][idx] <= 0:
                            if cv_step[-1][idx] > controller.real_cv[-1][idx]:
                                step = controller.real_cv
                            else:
                                step = cv_step

                        else:
                            step = cv_step
                    else:
                        if controller.real_cv[-1][idx]-controller.real_cv[-2][idx] <= 0:
                            if cv_step[-1][idx] > controller.real_cv[-1][idx]:
                                step = controller.real_cv
                            else:
                                step = cv_step
                        else:
                            step = controller.real_cv
                    # if cv_step[-1][idx]-cv_step[-2][idx] >= 0:
                    #     step = cv_step if cv_step[-1][idx] > controller.real_cv[-1][idx] else controller.real_cv
                    # else:
                    #     step = cv_step if cv_step[-1][idx] < controller.real_cv[-1][idx] else controller.real_cv
                    cv_ax.text(t_step[-1], step[-1][idx]+0.008*np.sign(step[-1][idx]-step[-2][idx]), plot_text)

            loss_ax.set_xlim([0, len(t_step)])
            loss_ax.set_ylim([0, loss_step[-1] + 1])
            int_loss_line.set_xdata(t_step)
            int_loss_line.set_ydata(loss_step)

            fig.canvas.draw()
            fig.canvas.flush_events()
    print('Final loss:', loss, 'Max step loss', max_step_loss, 'Negative loss', neg_loss)
    print(f'old loss: {old_loss}, sum loss + neg loss: {loss + neg_loss}')
    controller.close()
    # if do_plot:
    #     plt.show()

    results = {
        'total loss': loss,
        'negative loss': neg_loss,
        'max step loss': max_step_loss,
        'cv_step': cv_step,
        'x_step': x_step,
        'T_step': T_step,
        'T_opt_step': T_opt_step,
        't_step': t_step,
        'u_step': u_step,
        'd_step': d_step,
        'u_opt_step': u_opt_step,
        'dist_table': disturb_table,
    }

    return results


if __name__ == '__main__':
    from disturbance_tables import distrub_table2, distrub_table1

    disturb_table = distrub_table1

    # full_run(disturb_table)